import time
import torch
import pytest
from koi._runtime import lib, ffi
from koi.utils import void_ptr, show_diff_result, time_kernel, koi_tensor

dtype = torch.float16
dev = "cuda"

test_dims = (
    (64 * 1024, 4096, 512),
    (64 * 1024, 4096, 1536),
)
iters = 500


@pytest.mark.parametrize("dims", test_dims)
def test_tiled_mm_swiglu(dims):
    if lib.koi_tc_is_available(lib.KOI_F16) == lib.KOI_NOT_SUPPORTED:
        print("Koi tensor core kernels not available, skipping test")
        return

    M, N, K = dims
    stream = torch.cuda.default_stream()
    stream_ptr = ffi.cast("void *", stream.cuda_stream)

    with torch.no_grad():
        torch.manual_seed(42)

        input = torch.rand((M, K), device=dev, dtype=dtype) * 0.5 - 0.25
        weights = torch.rand((N, K), device=dev, dtype=dtype) * 0.5 - 0.25
        out_ref = torch.zeros((M, N // 2), device=dev, dtype=dtype)
        out_koi = torch.zeros((M // 16, N // 16, 16, 8), device=dev, dtype=dtype)

        def torch_impl():
            mm = (input @ weights.T).view((M, N // 2, 2))
            out_ref[:] = torch.nn.functional.silu(mm[:, :, 1]) * mm[:, :, 0]

        t_ref = time_kernel(iters, stream, torch_impl, None)

        input = input.view((M // 16, 16, K // 8, 8)).transpose(1, 2).contiguous()
        weights = (
            weights.view((N // 32, 16, 2, K // 8, 8))
            .permute((0, 2, 3, 1, 4))
            .contiguous()
            .view((N // 16, K // 8, 16, 8))
        )

        ctr_bfr = torch.zeros((iters + 1,), device=dev, dtype=torch.int32)
        ctr = 0

        def koi_impl():
            nonlocal ctr
            args = (
                stream_ptr,
                koi_tensor(input, ["M", "K", "m", "k"]),
                koi_tensor(weights, ["N", "K", "n", "k"]),
                koi_tensor(out_koi, ["M", "N", "m", "n"]),
                void_ptr(ctr_bfr[ctr]),
                1,
            )
            ctr += 1
            return lib.koi_mm_swiglu(*args)

        time_kernel(iters // 2, stream, koi_impl, lib.KOI_SUCCESS)
        ctr = 0
        ctr_bfr[:] = 0
        t_koi = time_kernel(iters, stream, koi_impl, lib.KOI_SUCCESS)

        ops = iters * M * N * K * 2
        tflops = 1e-12 * ops / t_koi

        out_koi2 = out_koi.permute((0, 2, 1, 3)).contiguous().view((M, N // 2))

        label = f"{dims}, ref {t_ref:.3f}s, koi {t_koi:.3f}s, {tflops:.3f} TFlops"
        show_diff_result(label, out_ref, out_koi2, 1e-2, 0.2, __name__ != "__main__")

        input[:] = 0
        weights[:] = 0
        ctr = 0
        ctr_bfr[:] = 0
        t_koi = time_kernel(iters, stream, koi_impl, lib.KOI_SUCCESS)
        tflops = 1e-12 * ops / t_koi
        print(f"All zero {t_koi:.3f}s, {tflops:.3f} TFlops")


if __name__ == "__main__":
    for test_dim in test_dims:
        test_tiled_mm_swiglu(test_dim)
